//            Copyright Daniel Trebbien 2010.
// Distributed under the Boost Software License, Version 1.0.
//   (See accompanying file LICENSE_1_0.txt or the copy at
//         http://www.boost.org/LICENSE_1_0.txt)

#ifndef BOOST_GRAPH_STOER_WAGNER_MINCUT_HPP
#define BOOST_GRAPH_STOER_WAGNER_MINCUT_HPP 1

#include <algorithm>
#include <functional>
#include <map>
#include <utility>
#include <vector>
#include <boost/concept_check.hpp>
#include <boost/scoped_array.hpp>
#include <boost/concept/assert.hpp>
#include <boost/graph/adjacency_list.hpp>
#include <boost/graph/exception.hpp>
#include <boost/graph/graph_traits.hpp>
#include <boost/graph/iteration_macros.hpp>
#include <boost/graph/named_function_params.hpp>
#include <boost/graph/detail/d_ary_heap.hpp>
#include <boost/property_map/property_map.hpp>
#include <boost/tuple/tuple.hpp>
#include <boost/typeof/typeof.hpp>

namespace boost {

namespace detail {

/**
 * \brief Performs a phase of the Stoer–Wagner min-cut algorithm
 *
 * Performs a \em phase of the Stoer–Wagner min-cut algorithm.
 *
 * As described by Stoer & Wagner (1997), a phase is simply a maximum adjacency search
 * (also called a maximum cardinality search), which results in the selection of two vertices
 * \em s and \em t, and, as a side product, a minimum <em>s</em>-<em>t</em> cut of
 * the input graph. Here, the input graph is basically \p g, but some vertices are virtually
 * assigned to others as a way of viewing \p g as a graph with some sets of
 * vertices merged together.
 *
 * This implementation is a translation of pseudocode by Professor Uri Zwick,
 * School of Computer Science, Tel Aviv University.
 *
 * \pre \p g is a connected, undirected graph
 * \param[in] g the input graph
 * \param[in] weights a property map from each edge to its weight (a non-negative value)
 * \param[in] assignments a property map from each vertex to the vertex that it is assigned to
 * \param[out] indicesInHeap a read/write property map from each vertex to its index in a priority queue.
 *     This map serves as work space, and no particular meaning should be derived from property values
 *     after completion of this function.
 * \param[out] wAs a read/write property map from each vertex to an accumulated weight value. This map
 *     serves as work space, and no particular meaning should be derived from property values
 *     after completion of this function.
 * \returns a tuple (\em s, \em t, \em w) of the "<em>s</em>" and "<em>t</em>"
 *     of the minimum <em>s</em>-<em>t</em> cut and the cut weight \em w
 *     of the minimum <em>s</em>-<em>t</em> cut.
 * \see http://www.cs.tau.ac.il/~zwick/grad-algo-08/gmc.pdf
 *
 * \author Daniel Trebbien
 * \date 2010-07-03
 */
template <class UndirectedGraph, class WeightMap, class VertexAssignmentMap, class IndexInHeapMap, class DistanceMap>
boost::tuple<typename boost::graph_traits<UndirectedGraph>::vertex_descriptor, typename boost::graph_traits<UndirectedGraph>::vertex_descriptor, typename boost::property_traits<WeightMap>::value_type> stoer_wagner_phase(const UndirectedGraph& g, WeightMap weights, VertexAssignmentMap assignments, IndexInHeapMap indicesInHeap, DistanceMap wAs)
{
    typedef typename boost::graph_traits<UndirectedGraph>::vertex_descriptor vertex_descriptor;
    typedef typename boost::property_traits<WeightMap>::value_type weight_type;
    
    boost::d_ary_heap_indirect<vertex_descriptor, 4, IndexInHeapMap, DistanceMap, std::greater<weight_type> > pq(wAs, indicesInHeap);
    
    BGL_FORALL_VERTICES_T(v, g, UndirectedGraph) {
        put(wAs, v, weight_type(0));
        
        if (v == get(assignments, v)) { // foreach u \in V do
            pq.push(v);
        }
    }
    
    vertex_descriptor s, t;
    weight_type w;
    while (!pq.empty()) { // while PQ \neq {} do
        const vertex_descriptor u = pq.top(); // u = extractmax(PQ)
        w = get(wAs, u);
        pq.pop();
        
        s = t; t = u;
        BGL_FORALL_VERTICES_T(uPrime, g, UndirectedGraph) {
            if (get(assignments, uPrime) == u) {
                BGL_FORALL_OUTEDGES_T(uPrime, e, g, UndirectedGraph) { // foreach (u, v) \in E do
                    const vertex_descriptor v = get(assignments, target(e, g));
                    
                    if (pq.contains(v)) { // if v \in PQ then
                        put(wAs, v, get(wAs, v) + get(weights, e));
                        pq.update(v);
                    }
                }
            }
        }
    }
    
    return boost::make_tuple(s, t, w);
}

} // end `namespace detail` within `namespace boost`

/**
 * \brief Computes a min-cut of the input graph
 *
 * Computes a min-cut of the input graph using the Stoer–Wagner algorithm.
 *
 * \pre \p g is a connected, undirected graph
 * \param[in] g the input graph
 * \param[in] weights a property map from each edge to its weight (a non-negative value)
 * \param[out] parities a writable property map from each vertex to a bool type object for
 *     distinguishing the two vertex sets of the min-cut
 * \param[out] assignments a read/write property map from each vertex to a \c vertex_descriptor object. This
 *     map serves as work space, and no particular meaning should be derived from property values
 *     after completion of the algorithm.
 * \param[out] indicesInHeap a read/write property map from each vertex to its index in a priority queue.
 *     This map serves as work space for <tt>boost::detail::stoer_wagner_phase</tt>,
 *     and no particular meaning should be derived from property values after completion of this function.
 * \param[out] wAs a read/write property map from each vertex to an accumulated weight value. This map
 *     serves as work space for <tt>boost::detail::stoer_wagner_phase</tt>,
 *     and no particular meaning should be derived from property values after completion of this function.
 * \returns the cut weight of the min-cut
 * \see http://citeseerx.ist.psu.edu/viewdoc/download?doi=10.1.1.114.6687&rep=rep1&type=pdf
 * \see http://citeseerx.ist.psu.edu/viewdoc/download?doi=10.1.1.31.614&rep=rep1&type=pdf
 *
 * \author Daniel Trebbien
 * \date 2010-07-03
 */
template <class UndirectedGraph, class WeightMap, class ParityMap, class VertexAssignmentMap, class IndexInHeapMap, class DistanceMap>
typename boost::property_traits<WeightMap>::value_type stoer_wagner_mincut(const UndirectedGraph& g, WeightMap weights, ParityMap parities, VertexAssignmentMap assignments, IndexInHeapMap indicesInHeap, DistanceMap wAs)
{
    BOOST_CONCEPT_ASSERT((boost::IncidenceGraphConcept<UndirectedGraph>));
    BOOST_CONCEPT_ASSERT((boost::VertexListGraphConcept<UndirectedGraph>));

    typedef typename boost::graph_traits<UndirectedGraph>::vertex_descriptor vertex_descriptor;
    typedef typename boost::graph_traits<UndirectedGraph>::vertices_size_type vertices_size_type;
    typedef typename boost::graph_traits<UndirectedGraph>::edge_descriptor edge_descriptor;
    
    BOOST_CONCEPT_ASSERT((boost::Convertible<typename boost::graph_traits<UndirectedGraph>::directed_category, boost::undirected_tag>));
    BOOST_CONCEPT_ASSERT((boost::ReadablePropertyMapConcept<WeightMap, edge_descriptor>));
    typedef typename boost::property_traits<WeightMap>::value_type weight_type;
    BOOST_CONCEPT_ASSERT((boost::WritablePropertyMapConcept<ParityMap, vertex_descriptor>));
    typedef typename boost::property_traits<ParityMap>::value_type parity_type;
    BOOST_CONCEPT_ASSERT((boost::Convertible<bool, parity_type>));
    BOOST_CONCEPT_ASSERT((boost::ReadWritePropertyMapConcept<VertexAssignmentMap, vertex_descriptor>));
    BOOST_CONCEPT_ASSERT((boost::Convertible<vertex_descriptor, typename boost::property_traits<VertexAssignmentMap>::value_type>));
    BOOST_CONCEPT_ASSERT((boost::ReadWritePropertyMapConcept<IndexInHeapMap, vertex_descriptor>));
    BOOST_CONCEPT_ASSERT((boost::ReadWritePropertyMapConcept<DistanceMap, vertex_descriptor>));
    
    vertices_size_type n = num_vertices(g);
    if (n < 2)
        throw boost::bad_graph("the input graph must have at least two vertices.");
    
    // initialize `assignments` (all vertices are initially assigned to themselves)
    BGL_FORALL_VERTICES_T(v, g, UndirectedGraph) {
        put(assignments, v, v);
    }
    
    vertex_descriptor s, t;
    weight_type bestW;
    
    boost::tie(s, t, bestW) = boost::detail::stoer_wagner_phase(g, weights, assignments, indicesInHeap, wAs);
    BGL_FORALL_VERTICES_T(v, g, UndirectedGraph) {
        put(parities, v, parity_type(v == t ? true : false));
    }
    put(assignments, t, s);
    --n;
    
    for (; n >= 2; --n) {
        weight_type w;
        boost::tie(s, t, w) = boost::detail::stoer_wagner_phase(g, weights, assignments, indicesInHeap, wAs);
        
        if (w < bestW) {
            BGL_FORALL_VERTICES_T(v, g, UndirectedGraph) {
                put(parities, v, parity_type(get(assignments, v) == t ? true : false));
                
                if (get(assignments, v) == t) // all vertices that were assigned to t are now assigned to s
                    put(assignments, v, s);
            }
            
            bestW = w;
        }
        else {
            BGL_FORALL_VERTICES_T(v, g, UndirectedGraph) {
                if (get(assignments, v) == t) // all vertices that were assigned to t are now assigned to s
                    put(assignments, v, s);
            }
        }
        put(assignments, t, s);
    }

    return bestW;
}

namespace detail {

template <class Graph, class ArgPack>
struct vertex_assignment_map_maker {
  typedef typename boost::graph_traits<Graph>::vertex_descriptor vertex_descriptor;
  BOOST_STATIC_CONSTANT(
    bool,
    hasVertexAssignmentMap =
      (parameter_exists<ArgPack, boost::graph::keywords::tag::vertex_assignment_map>
       ::value));
  typedef color_map_maker_helper<hasVertexAssignmentMap, Graph, ArgPack, vertex_descriptor,
                                 typename boost::remove_const<
                                   typename boost::parameter::value_type<
                                              ArgPack,
                                              boost::graph::keywords::tag::vertex_assignment_map,
                                              vertex_descriptor
                                            >::type
                                          >::type> helper;
  typedef typename helper::map_type map_type;
  static inline map_type make_map(const Graph& g, const ArgPack& ap) {
    return helper::make_map(g, *vertices(g).first, ap[boost::graph::keywords::_vertex_assignment_map | 0], ap);
  }
};

template <class Graph, class ArgPack>
struct index_in_heap_map_maker {
  typedef typename boost::graph_traits<Graph>::vertex_descriptor vertex_descriptor;
  typedef typename std::vector<vertex_descriptor>::size_type heap_container_size_type;
  BOOST_STATIC_CONSTANT(
    bool,
    hasIndexInHeapMap =
      (parameter_exists<ArgPack, boost::graph::keywords::tag::index_in_heap_map>
       ::value));
  typedef color_map_maker_helper<hasIndexInHeapMap, Graph, ArgPack, heap_container_size_type,
                                 typename boost::remove_const<
                                   typename boost::parameter::value_type<
                                              ArgPack,
                                              boost::graph::keywords::tag::index_in_heap_map,
                                              heap_container_size_type
                                            >::type
                                          >::type> helper;
  typedef typename helper::map_type map_type;
  static inline map_type make_map(const Graph& g, const ArgPack& ap) {
    return helper::make_map(g, heap_container_size_type(-1), ap[boost::graph::keywords::_index_in_heap_map | 0], ap);
  }
};

template <class Graph, class WeightMap, class ArgPack>
struct distance_map_maker {
  typedef typename boost::property_traits<WeightMap>::value_type weight_type;
  BOOST_STATIC_CONSTANT(
    bool,
    hasDistanceMap =
      (parameter_exists<ArgPack, boost::graph::keywords::tag::distance_map>
       ::value));
  typedef color_map_maker_helper<hasDistanceMap, Graph, ArgPack, weight_type,
                                 typename boost::remove_const<
                                   typename boost::parameter::value_type<
                                              ArgPack,
                                              boost::graph::keywords::tag::distance_map,
                                              weight_type
                                            >::type
                                          >::type> helper;
  typedef typename helper::map_type map_type;
  static inline map_type make_map(const Graph& g, const ArgPack& ap) {
    return helper::make_map(g, weight_type(0), ap[boost::graph::keywords::_index_in_heap_map | 0], ap);
  }
};

} // end `namespace detail` within `namespace boost`

template <class UndirectedGraph, class WeightMap, class P, class T, class R>
inline typename boost::property_traits<WeightMap>::value_type stoer_wagner_mincut(const UndirectedGraph& g, WeightMap weights, const boost::bgl_named_params<P, T, R>& params)
{
    typedef boost::bgl_named_params<P, T, R> params_type;
    BOOST_GRAPH_DECLARE_CONVERTED_PARAMETERS(params_type, params);
    return boost::stoer_wagner_mincut(g,
               weights,
               choose_param(get_param(params, boost::parity_map_t()), boost::dummy_property_map()),
               boost::detail::vertex_assignment_map_maker<UndirectedGraph, arg_pack_type>::make_map(g, arg_pack),
               boost::detail::index_in_heap_map_maker<UndirectedGraph, arg_pack_type>::make_map(g, arg_pack),
               boost::detail::distance_map_maker<UndirectedGraph, WeightMap, arg_pack_type>::make_map(g, arg_pack)
           );
}

template <class UndirectedGraph, class WeightMap>
inline typename boost::property_traits<WeightMap>::value_type stoer_wagner_mincut(const UndirectedGraph& g, WeightMap weights)
{
    return boost::stoer_wagner_mincut(g, weights, boost::vertex_index_map(get(boost::vertex_index, g)));
}

} // end `namespace boost`

#include <boost/graph/iteration_macros_undef.hpp>

#endif // !BOOST_GRAPH_STOER_WAGNER_MINCUT_HPP
